// ******************************************************
// Functions for atlas specification files
// MRFSEG + GAMIXTURE software bundle: Voxel classification in volumetric 
// images using genetic algorithms and Markov random fields.
// The genetic algorithm is described in: 
// [1]  J. Tohka , E. Krestyannikov, I.D. Dinov , A. MacKenzie-Graham,
// D.W. Shattuck , U. Ruotsalainen, and A.W. Toga. 
// Genetic algorithms for finite mixture model based voxel classification
// in neuroimaging.  
// IEEE Transactions on Medical Imaging  , 26(5):696 - 711, 2007.
// (C) Jussi Tohka, 2005 - 2007 
// Laboratory of NeuroImaging, Department of Neurology, UCLA Medical School, CA, USA
// Institute of Signal Processing, Tampere University of Technology, Finland  
// Version 1.1
// e-mail jussi.tohka@tut.fi
// *****************************************************
// ******************************************************
// Permission to use, copy, modify, and distribute this software 
// for any purpose and without fee is hereby
// granted, provided that the above copyright notice appears in all
// copies.  The author, Tampere University of Technology and University of
// California, Los Angeles make no representations
// about the suitability of this software for any purpose.  It is
// provided "as is" without express or implied warranty.
// *****************************************************





#ifndef ATLASSPEC_H
#define ATLASSPEC_H

#include "analyze.h"

#define MAX_NAME_LEN 256
#define MAX_LABELS 20 
#define PURELABEL  0
#define PVELABEL 1

struct LabelType {
  bool pureLabel;
  int mixed[2];  
};

struct LabelList {
  int len;
  int list[MAX_LABELS];
};

struct AtlasSpec {
  int n; // Gives the number of different fuzzy regions
  int numberOfLabels; // Gives the number of possible labels, background included 
   
  char** labelnames;    // Gives the names of labels; 0 is reserved for background
  char** filenames;    // Gives the names of the files containg fuzzy masks
  char** regionnames;  // Gives the description of each region (optional)
  LabelList* permittedLabels; // Labels permitted for each region 
                       
  float** mrfConstants; // pairwise interactions in the mrf
  LabelType* labelTypes;  // types of labels (pve or gaussian)
  float** regionLowProb;
  float** regionUpProb;                  
};


//////////////////////////////////
// the atlas file should look like this:
// lines starting with # are for comments
//////////////////////////////////
// ATLAS NAME
// n numberOfLabels
// Region1name Region1filename LabelNOTPermitted1 2 3 ...
// Region2name Region2filename LabelNOTPermitted1 2 3 ...
//  ....
// Regionnname Regionnfilename
// Label1name Label1type1 Label1type2 Label1type3
// ...
// Labelnname Labelntype1 ....  
// Mrf matrix (numberofLabels + 1) x (numberOfLabels + 1)
// should be symmetric


bool readAtlasSpec(AtlasSpec* atlas,char* filename);
void freeAtlas(AtlasSpec* atlas);
int readAtlasImages(AtlasSpec* atlas,AnalyzeImage** atlasImages); 
void freeAtlasImages(AtlasSpec* atlas,AnalyzeImage** atlasImages);
bool maskAtlas(AtlasSpec* atlas, AnalyzeImage** atlasImages,AnalyzeImage* mask);


inline void constructLabelList(LabelList* labels,int maxLabel,int* remThese, int howManyToRemove) 
{
  int i,j,k;
  bool addLabel;

  k = 0;
  for(i = 1;i < (maxLabel + 1);i++) {
    addLabel = true;
    for(j = 0;j < howManyToRemove;j++) {
      if(remThese[j] == i) addLabel = false;
    }
    if(addLabel) {
      labels->list[k] = i;
      k++;
    }
  }
  labels->len = k;
} 

// inline void invertLabelList(LabelList* newlabels,LabelList* labels,int maxLabel, bool inclBG)
// {}

inline void printLabelInfo(AtlasSpec* atlas) 
{
  int i,j;

  cout << "Number of regions:" << atlas->n << " Number of Labels:" << atlas->numberOfLabels << endl;
  cout << "Label:   Name   Pure   compos " << endl;
  for(i = 0; i < atlas->numberOfLabels;i++) {
    cout << i << "        " << atlas->labelnames[i] << "   " << atlas->labelTypes[i].pureLabel << "   ";
    if(atlas->labelTypes[i].pureLabel) cout << endl;
    else cout <<  atlas->labelTypes[i].mixed[0] << " " <<  atlas->labelTypes[i].mixed[1] << endl;   
  }
  cout << "MRF" << endl;
  
  for(i = 0; i < atlas->numberOfLabels;i++) { 
    for(j = 0; j < atlas->numberOfLabels;j++) {
      cout.width(4); 
      cout << atlas->mrfConstants[i][j];
    }
    cout << endl;
  } 
  
}
#endif

